from .scvi import SCVI
from .totalvi import TOTALVI
from .linear_scvi import LinearSCVI
from .autozi import AUTOZI
from .scanvi import SCANVI
from .gimvi import GIMVI

__all__ = ["SCVI", "TOTALVI", "LinearSCVI", "AUTOZI", "SCANVI", "GIMVI"]
